import pytest

from aceai.core.base import AgentBase
from aceai.agent.session import SessionRecorder, SessionState, SessionStore
from aceai.llm import LLMResponse
from aceai.core.run_state import ToolRunState
from aceai.llm.models import LLMToolCall, LLMUsage
from aceai.llm.models import LLMStreamEvent
from aceai.agent.tui.events import TUIEvent
from aceai.agent.tui.session_adapter import tui_event_to_session_event
from aceai.agent.tui.session_replay import event_log_to_tui_events
from aceai.agent.tui.config import AceAITUIConfig
from aceai.agent.config import clear_config, current_config
from aceai.agent.tui import app as tui_app_module
from aceai.agent.tui.app import AceAITUI
from aceai.agent.tui.runner import AceAIConfiguredTUI, AceAIInteractiveTUI, AceAILiveTUI
from aceai.agent.tui.setup import (
    ConfigScreen,
    ConfigSelection,
    SkillConfigItem,
    ToolPermissionItem,
)
from aceai.agent.tui.widgets import ApprovalWidget
from aceai.agent.tui.widgets import CommandInput, StatusBarWidget
from textual.widgets import Button, Checkbox, Input, RichLog, Select, Static, TabbedContent


@pytest.fixture(autouse=True)
def tui_session_store(monkeypatch, tmp_path) -> SessionStore:
    store = SessionStore(tmp_path / "sessions")
    monkeypatch.setattr(tui_app_module, "SessionStore", lambda: store)
    return store


class StubExecutor:
    def select_tools(
        self,
        include: set[str] | None = None,
        exclude: set[str] | None = None,
    ) -> list[object]:
        if include and exclude:
            raise ValueError("Cannot specify both include and exclude")
        return []


class ApprovalExecutor:
    def __init__(self) -> None:
        self.calls: list[LLMToolCall] = []

    def select_tools(
        self,
        include: set[str] | None = None,
        exclude: set[str] | None = None,
    ) -> list[object]:
        if include and exclude:
            raise ValueError("Cannot specify both include and exclude")
        return []

    def resolve_invocation(self, tool_call: LLMToolCall):
        return ApprovalInvocation(tool_call)

    async def execute(
        self,
        invocation,
        *,
        tool_state: ToolRunState,
    ) -> str:
        self.calls.append(invocation.call)
        return '{"ok":true}'


class ApprovalInvocation:
    def __init__(self, call: LLMToolCall) -> None:
        self.call = call
        self.approval_required = True
        self.tool = ApprovalTool(call.name)


class ApprovalTool:
    def __init__(self, name: str) -> None:
        self.name = name
        self.metadata = ApprovalToolMetadata()


class ApprovalToolMetadata:
    require_approval = True
    approval_policy = "filesystem_write"


class StubLLMService:
    def __init__(self, events: list[LLMStreamEvent]) -> None:
        self._events = list(events)
        self.calls: list[dict] = []

    async def stream(self, **request):
        self.calls.append(request)
        for event in self._events:
            yield event

    async def complete(self, **request) -> LLMResponse:
        raise AssertionError("live TUI should use streaming")


class MultiRunLLMService:
    def __init__(self, streams: list[list[LLMStreamEvent]]) -> None:
        self._streams = [list(stream) for stream in streams]
        self.calls: list[dict] = []

    async def stream(self, **request):
        self.calls.append(request)
        events = self._streams.pop(0)
        for event in events:
            yield event

    async def complete(self, **request) -> LLMResponse:
        raise AssertionError("live TUI should use streaming")


@pytest.mark.anyio
async def test_live_tui_streams_agent_events_into_state() -> None:
    llm_service = StubLLMService(
        [
            LLMStreamEvent(
                event_type="response.output_text.delta",
                text_delta="hello",
            ),
            LLMStreamEvent(
                event_type="response.completed",
                response=LLMResponse(text="hello"),
            ),
        ]
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAILiveTUI(agent, "Question?")

    async with app.run_test() as pilot:
        await pilot.pause(0.1)
        assert app._state.status == "completed"
        assert app._state.final_answer == "hello"
        assert llm_service.calls[0]["messages"][-1].content[0]["data"] == "Question?"


@pytest.mark.anyio
async def test_interactive_tui_submits_question_from_input(
    tui_session_store: SessionStore,
) -> None:
    llm_service = StubLLMService(
        [
            LLMStreamEvent(
                event_type="response.output_text.delta",
                text_delta="answer",
            ),
            LLMStreamEvent(
                event_type="response.completed",
                response=LLMResponse(text="answer"),
            ),
        ]
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)
    assert app._session_recorder is None
    assert app._session_id is None

    async with app.run_test() as pilot:
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, "What now?"))
        await pilot.pause(0.1)

        assert app._state.status == "completed"
        assert app._state.final_answer == "answer"
        assert command_input.value == ""
        assert llm_service.calls[0]["messages"][-1].content[0]["data"] == "What now?"
        assert app._session_recorder is not None
        assert app._session_id is not None
        event_log = tui_session_store.load_event_log(app._session_id)
        assert event_log.events[0].payload["content"] == "What now?"
        run_ids = {session_event.run_id for session_event in event_log.events}
        assert len(run_ids) == 1
        assert "" not in run_ids
        assert event_log.get_run(next(iter(run_ids))).question == "What now?"


@pytest.mark.anyio
async def test_interactive_tui_clear_command_resets_state() -> None:
    llm_service = StubLLMService([])
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test():
        app.load_events([])
        app._state = app._state.__class__(
            status="completed",
            final_answer="done",
        )
        command_input = app.query_one(CommandInput)
        command_input.value = "/clear"
        app.on_input_submitted(Input.Submitted(command_input, "/clear"))

        assert app._state.status == "idle"
        assert app._state.events == []
        assert command_input.value == ""


@pytest.mark.anyio
async def test_interactive_tui_keeps_history_between_questions() -> None:
    llm_service = MultiRunLLMService(
        [
            [
                LLMStreamEvent(
                    event_type="response.output_text.delta",
                    text_delta="first",
                ),
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="first"),
                ),
            ],
            [
                LLMStreamEvent(
                    event_type="response.output_text.delta",
                    text_delta="second",
                ),
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="second"),
                ),
            ],
        ]
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test() as pilot:
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, "First?"))
        await pilot.pause(0.1)
        first_event_count = len(app._state.events)

        app.on_input_submitted(Input.Submitted(command_input, "Second?"))
        await pilot.pause(0.1)

        assert len(app._state.events) > first_event_count
        assert app._state.final_answer == "second"
        assert llm_service.calls[0]["messages"][-1].content[0]["data"] == "First?"
        assert llm_service.calls[1]["messages"][-1].content[0]["data"] == "Second?"


@pytest.mark.anyio
async def test_interactive_tui_approves_suspended_tool_and_continues() -> None:
    call = LLMToolCall(
        name="write_text_file",
        arguments='{"path":"x","content":"hello"}',
        call_id="call-1",
    )
    llm_service = MultiRunLLMService(
        [
            [
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="use tool", tool_calls=[call]),
                ),
            ],
            [
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="done"),
                ),
            ],
        ]
    )
    executor = ApprovalExecutor()
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=executor,  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test() as pilot:
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, "Write it"))
        await pilot.pause(0.1)

        assert app._state.status == "suspended"
        assert executor.calls == []
        assert app._active_runtime is not None
        assert app._active_runtime.status == "suspended"
        assert command_input.placeholder == "Choose Approve or Reject"
        status = app.query_one(StatusBarWidget)
        assert "action: choose Approve or Reject" in status.current_text
        approval = app.query_one(ApprovalWidget)
        assert not approval.has_class("collapsed")
        assert approval.query_one("#approval-approve", Button).label.plain == "A Approve"
        assert approval.query_one("#approval-reject", Button).label.plain == "R Reject"
        assert "content:" in str(approval.query_one("#approval-summary", Static).render())

        approval.post_message(ApprovalWidget.Selected(approved=True))
        await pilot.pause(0.1)

        assert app._state.status == "completed"
        assert app._state.final_answer == "done"
        assert command_input.placeholder == "Ask AceAI or type /quit"
        assert approval.has_class("collapsed")
        assert executor.calls == [call]
        assert app._llm_history[-1].role == "assistant"
        assert app._llm_history[-1].content[0]["data"] == "done"


@pytest.mark.anyio
async def test_interactive_tui_approves_suspended_tool_with_keyboard() -> None:
    call = LLMToolCall(
        name="write_text_file",
        arguments='{"path":"x","content":"hello"}',
        call_id="call-1",
    )
    llm_service = MultiRunLLMService(
        [
            [
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="use tool", tool_calls=[call]),
                ),
            ],
            [
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="done"),
                ),
            ],
        ]
    )
    executor = ApprovalExecutor()
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=executor,  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test() as pilot:
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, "Write it"))
        await pilot.pause(0.1)

        await pilot.press("a")
        await pilot.pause(0.1)

        assert app._state.status == "completed"
        assert app._state.final_answer == "done"
        assert executor.calls == [call]


@pytest.mark.anyio
async def test_interactive_tui_shows_next_approval_after_resume_suspends_again() -> None:
    first_call = LLMToolCall(
        name="write_text_file",
        arguments='{"path":"x","content":"hello"}',
        call_id="call-1",
    )
    second_call = LLMToolCall(
        name="run_shell_command",
        arguments='{"command":"python binary_search.py"}',
        call_id="call-2",
    )
    llm_service = MultiRunLLMService(
        [
            [
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="write file", tool_calls=[first_call]),
                ),
            ],
            [
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="run file", tool_calls=[second_call]),
                ),
            ],
        ]
    )
    executor = ApprovalExecutor()
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=executor,  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test() as pilot:
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, "Write and run it"))
        await pilot.pause(0.1)

        approval = app.query_one(ApprovalWidget)
        assert app._state.status == "suspended"
        assert not approval.has_class("collapsed")
        assert "write_text_file" in str(
            approval.query_one("#approval-summary", Static).render()
        )

        approval.post_message(ApprovalWidget.Selected(approved=True))
        await pilot.pause(0.1)

        assert app._state.status == "suspended"
        assert app._active_runtime is not None
        assert app._active_runtime.status == "suspended"
        assert command_input.placeholder == "Choose Approve or Reject"
        assert executor.calls == [first_call]
        assert not approval.has_class("collapsed")
        summary = str(approval.query_one("#approval-summary", Static).render())
        assert "run_shell_command" in summary
        assert "python binary_search.py" in summary


@pytest.mark.anyio
async def test_interactive_tui_model_command_updates_next_request_metadata() -> None:
    llm_service = StubLLMService(
        [
            LLMStreamEvent(
                event_type="response.output_text.delta",
                text_delta="answer",
            ),
            LLMStreamEvent(
                event_type="response.completed",
                response=LLMResponse(text="answer"),
            ),
        ]
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test() as pilot:
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, "/model gpt-5.5"))
        app.on_input_submitted(Input.Submitted(command_input, "What now?"))
        await pilot.pause(0.1)

        assert app._selected_model == "gpt-5.5"
        assert app._status_model == "gpt-5.5"
        assert llm_service.calls[0]["metadata"]["model"] == "gpt-5.5"


@pytest.mark.anyio
async def test_interactive_tui_persists_selected_model_in_session_state(tmp_path) -> None:
    store = SessionStore(tmp_path)
    metadata = store.create_session()
    SessionRecorder(store, metadata.session_id).record(
        tui_event_to_session_event(TUIEvent.user_message("keep this session"))
    )
    llm_service = StubLLMService([])
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(
        agent,
        session_recorder=SessionRecorder(store, metadata.session_id),
        session_id=metadata.session_id,
    )

    async with app.run_test():
        app.switch_model("gpt-5.5")

    assert store.get_session_state(metadata.session_id) == SessionState(
        selected_provider="openai",
        selected_model="gpt-5.5",
    )


@pytest.mark.anyio
async def test_interactive_tui_model_only_session_finalizes_as_empty(tmp_path) -> None:
    store = SessionStore(tmp_path)
    metadata = store.create_session()
    llm_service = StubLLMService([])
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(
        agent,
        session_recorder=SessionRecorder(store, metadata.session_id),
        session_id=metadata.session_id,
    )

    async with app.run_test():
        app.append_event(TUIEvent.session_notice(f"Resumed session {metadata.session_id}"))
        app.switch_model("gpt-5.5")

    assert store.list_sessions() == []


@pytest.mark.anyio
async def test_interactive_tui_status_bar_shows_selected_model() -> None:
    llm_service = StubLLMService([])
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test():
        status = app.query_one(StatusBarWidget)

        assert "model: gpt-4o" in status.current_text

        app.switch_model("gpt-5.5")

        assert "model: gpt-5.5" in status.current_text


@pytest.mark.anyio
async def test_interactive_tui_status_bar_shows_usage() -> None:
    llm_service = StubLLMService(
        [
            LLMStreamEvent(
                event_type="response.completed",
                response=LLMResponse(
                    text="answer",
                    model="gpt-5.5",
                    usage=LLMUsage(
                        input_tokens=1_200,
                        cached_input_tokens=200,
                        output_tokens=300,
                        total_tokens=1_500,
                    ),
                ),
            ),
        ]
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test() as pilot:
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, "What now?"))
        await pilot.pause(0.1)

        status = app.query_one(StatusBarWidget)
        assert "ctx: 1,200" in status.current_text
        assert "cost: $0.0141" in status.current_text


@pytest.mark.anyio
async def test_interactive_tui_metadata_lists_runtime_usage_and_skills(tmp_path) -> None:
    skill_dir = tmp_path / "skills" / "debugger"
    skill_dir.mkdir(parents=True)
    (skill_dir / "SKILL.md").write_text(
        "---\n"
        "name: debugger\n"
        "description: Debug flaky tests.\n"
        "---\n"
        "# Debugger\n",
        encoding="utf-8",
    )
    llm_service = StubLLMService([])
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
        skill_path=tmp_path / "skills",
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test():
        sections = app._metadata_sections()

    section_lines = {
        section.title: "\n".join(section.lines)
        for section in sections
    }
    assert "model: gpt-4o" in section_lines["Runtime"]
    assert "session cost: -" in section_lines["Usage"]
    assert "provider: openai" in section_lines["Agent"]
    assert "debugger: Debug flaky tests." in section_lines["Skills (1)"]


@pytest.mark.anyio
async def test_metadata_screen_scrolls_and_keeps_close_button(tmp_path) -> None:
    skill_root = tmp_path / "skills"
    for index in range(20):
        skill_dir = skill_root / f"skill_{index}"
        skill_dir.mkdir(parents=True)
        (skill_dir / "SKILL.md").write_text(
            "---\n"
            f"name: skill_{index}\n"
            f"description: Skill number {index}.\n"
            "---\n"
            "# Skill\n",
            encoding="utf-8",
        )
    llm_service = StubLLMService([])
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
        skill_path=skill_root,
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test(size=(100, 24)) as pilot:
        app.open_metadata_screen()
        await pilot.pause(0.1)
        body = app.screen.query_one("#metadata-body", RichLog)
        close_button = app.screen.query_one("#metadata-close", Button)

        assert body.max_scroll_y > 0
        assert close_button.display

        await pilot.press("pagedown")

        assert body.scroll_y > 0


@pytest.mark.anyio
async def test_interactive_tui_model_selection_callback_updates_model() -> None:
    llm_service = StubLLMService(
        [
            LLMStreamEvent(
                event_type="response.completed",
                response=LLMResponse(text="answer"),
            ),
        ]
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test() as pilot:
        command_input = app.query_one(CommandInput)
        app._handle_config_selection("gpt-5.5")
        app.on_input_submitted(Input.Submitted(command_input, "Use selected model"))
        await pilot.pause(0.1)

        assert app._selected_model == "gpt-5.5"
        assert llm_service.calls[0]["metadata"]["model"] == "gpt-5.5"


@pytest.mark.anyio
async def test_interactive_tui_c_key_opens_config_screen() -> None:
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=StubLLMService([]),  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)
    calls: list[str] = []
    app.open_config_screen = lambda: calls.append("config")

    async with app.run_test() as pilot:
        await pilot.press("c")

    assert calls == ["config"]


@pytest.mark.anyio
async def test_interactive_tui_config_command_opens_config_screen() -> None:
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=StubLLMService([]),  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)
    calls: list[str] = []
    app.open_config_screen = lambda: calls.append("config")

    async with app.run_test():
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, "/config"))

    assert calls == ["config"]


@pytest.mark.anyio
async def test_interactive_tui_info_command_uses_registered_alias() -> None:
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=StubLLMService([]),  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)
    calls: list[str] = []
    app.open_metadata_screen = lambda: calls.append("metadata")

    async with app.run_test():
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, "/info"))

    assert calls == ["metadata"]


@pytest.mark.anyio
async def test_interactive_tui_resume_command_uses_registered_arg_handler(
    tmp_path,
) -> None:
    store = SessionStore(tmp_path)
    first = store.create_session()
    second = store.create_session()
    SessionRecorder(store, first.session_id).record(
        tui_event_to_session_event(TUIEvent.user_message("first"))
    )
    SessionRecorder(store, second.session_id).record(
        tui_event_to_session_event(TUIEvent.user_message("second"))
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=StubLLMService([]),  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(
        agent,
        initial_events=event_log_to_tui_events(store.load_event_log(first.session_id)),
        initial_history=[],
        session_recorder=SessionRecorder(store, first.session_id),
        session_id=first.session_id,
    )

    async with app.run_test():
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, f"/resume {second.session_id}"))

    assert app._session_id == second.session_id
    assert app._state.events[0].content == "second"


@pytest.mark.anyio
async def test_interactive_tui_unknown_slash_input_runs_as_question() -> None:
    llm_service = StubLLMService(
        [
            LLMStreamEvent(
                event_type="response.completed",
                response=LLMResponse(text="answer"),
            ),
        ]
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test() as pilot:
        command_input = app.query_one(CommandInput)
        app.on_input_submitted(Input.Submitted(command_input, "/unknown command"))
        await pilot.pause(0.1)

    assert llm_service.calls[0]["messages"][-1].content[0]["data"] == "/unknown command"


@pytest.mark.anyio
async def test_interactive_tui_returns_focus_to_stream_after_submit() -> None:
    llm_service = StubLLMService(
        [
            LLMStreamEvent(
                event_type="response.completed",
                response=LLMResponse(text="answer"),
            ),
        ]
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(agent)

    async with app.run_test() as pilot:
        command_input = app.query_one(CommandInput)
        command_input.focus()
        app.on_input_submitted(Input.Submitted(command_input, "Use selected model"))
        await pilot.pause(0.1)

        assert not command_input.has_focus
        assert app.query_one("#stream").has_focus


@pytest.mark.anyio
async def test_config_screen_prefills_masked_api_key() -> None:
    screen = ConfigScreen(
        provider_name="openai",
        current_model="gpt-5.5",
        default_model="gpt-5.5",
        skills="auto",
        api_keys={"openai": "sk-test-ending"},
    )

    async with AceAITUI([]).run_test() as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()

        assert screen.query_one("#api-key", Input).value == "*****************ding"


@pytest.mark.anyio
async def test_config_screen_apply_restores_masked_api_key() -> None:
    screen = ConfigScreen(
        provider_name="openai",
        current_model="gpt-5.5",
        default_model="gpt-5.5",
        skills="auto",
        api_keys={"openai": "sk-test-ending"},
    )

    async with AceAITUI([]).run_test() as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()
        selections: list[ConfigSelection | None] = []

        def dismiss(selection: ConfigSelection | None) -> None:
            selections.append(selection)

        screen.dismiss = dismiss
        _press_config_apply(screen)

        assert selections == [
            ConfigSelection(
                provider="openai",
                model="gpt-5.5",
                default_model="gpt-5.5",
                api_key="sk-test-ending",
                skills="auto",
                skill_selection_mode="selected",
                enabled_skills=(),
            )
        ]


@pytest.mark.anyio
async def test_config_screen_uses_model_as_default_model_and_separates_skills() -> None:
    skill_item = SkillConfigItem(
        name="developer",
        description=(
            "Practical software development workflow with careful repository "
            "inspection, implementation, and verification."
        ),
        location="/skills/developer/SKILL.md",
    )
    screen = ConfigScreen(
        provider_name="openai",
        current_model="gpt-5.5",
        default_model="gpt-4o",
        skills="auto",
        skill_items=(skill_item,),
        skill_selection_mode="all",
        enabled_skills=(),
        api_keys={"openai": "sk-test-ending"},
    )

    async with AceAITUI([]).run_test() as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()
        selections: list[ConfigSelection | None] = []

        def dismiss(selection: ConfigSelection | None) -> None:
            selections.append(selection)

        screen.dismiss = dismiss
        screen.query_one("#model", Input).value = "gpt-4o"
        screen_ids = {node.id for node in screen.query("*") if node.id is not None}
        _press_config_apply(screen)

        assert "default-model" not in screen_ids
        assert screen.query_one("#api-key", Input).region.y < screen.query_one(
            "#config-skills-list"
        ).region.y
        assert str(screen.query_one("#skill-0", Checkbox).label) == "developer"
        assert (
            "Practical software development workflow"
            in str(screen.query_one("#skill-description-0", Static).render())
        )
        assert selections == [
            ConfigSelection(
                provider="openai",
                model="gpt-4o",
                default_model="gpt-4o",
                api_key="sk-test-ending",
                skills="auto",
                skill_selection_mode="selected",
                enabled_skills=("developer",),
            )
        ]


@pytest.mark.anyio
async def test_config_screen_is_fullscreen_and_splits_system_prompt_tab() -> None:
    screen = ConfigScreen(
        provider_name="openai",
        current_model="gpt-5.5",
        default_model="gpt-5.5",
        skills="auto",
        api_keys={"openai": "sk-test-ending"},
        system_prompt="You are AceAI.\n\nSkill instructions are active.",
    )

    async with AceAITUI([]).run_test(size=(100, 30)) as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()

        panel = screen.query_one("#config-panel")
        tabs = screen.query_one("#config-tabs", TabbedContent)
        settings_tab = screen.query_one("#settings-tab")
        prompt_tab = screen.query_one("#system-prompt-tab")
        prompt = screen.query_one("#system-prompt", Static)

        assert panel.region.width == 100
        assert panel.region.height == 30
        assert tabs.active == "settings-tab"
        assert screen.query_one("#provider", Input) in settings_tab.query("*")
        assert prompt in prompt_tab.query("*")
        assert "You are AceAI" in str(prompt.render())
        assert "Skill instructions are active" in str(prompt.render())


@pytest.mark.anyio
async def test_config_screen_has_tool_permissions_tab_and_selects_policy() -> None:
    tool_items = (
        ToolPermissionItem(
            name="run_shell_command",
            description="Run a shell command.",
            permission="ask",
        ),
        ToolPermissionItem(
            name="read_text_file",
            description="Read a file.",
            permission="always",
        ),
    )
    screen = ConfigScreen(
        provider_name="openai",
        current_model="gpt-5.5",
        default_model="gpt-5.5",
        skills="auto",
        api_keys={"openai": "sk-test-ending"},
        tool_permission_items=tool_items,
    )

    async with AceAITUI([]).run_test() as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()
        selections: list[ConfigSelection | None] = []

        def dismiss(selection: ConfigSelection | None) -> None:
            selections.append(selection)

        screen.dismiss = dismiss
        tabs = screen.query_one("#config-tabs", TabbedContent)
        tool_select = screen.query_one("#tool-permission-0", Select)

        assert screen.query_one("#tool-permissions-tab") in tabs.query("*")
        assert tool_select.value == "ask"

        tool_select.value = "never"
        _press_config_apply(screen)

        assert selections[-1].tool_permissions == {
            "run_shell_command": "never",
            "read_text_file": "always",
        }


@pytest.mark.anyio
async def test_config_screen_can_allow_or_disable_all_tools() -> None:
    tool_items = (
        ToolPermissionItem(
            name="run_shell_command",
            description="Run a shell command.",
            permission="ask",
        ),
        ToolPermissionItem(
            name="read_text_file",
            description="Read a file.",
            permission="always",
        ),
    )
    screen = ConfigScreen(
        provider_name="openai",
        current_model="gpt-5.5",
        default_model="gpt-5.5",
        skills="auto",
        api_keys={"openai": "sk-test-ending"},
        tool_permission_items=tool_items,
    )

    async with AceAITUI([]).run_test() as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()
        selections: list[ConfigSelection | None] = []

        def dismiss(selection: ConfigSelection | None) -> None:
            selections.append(selection)

        screen.dismiss = dismiss

        class DisablePressed:
            button = screen.query_one("#disable-all-tools", Button)

        screen.on_button_pressed(DisablePressed())
        assert screen.query_one("#tool-permission-0", Select).value == "never"
        assert screen.query_one("#tool-permission-1", Select).value == "never"

        class AllowPressed:
            button = screen.query_one("#allow-all-tools", Button)

        screen.on_button_pressed(AllowPressed())
        _press_config_apply(screen)

        assert selections[-1].tool_permissions == {
            "run_shell_command": "always",
            "read_text_file": "always",
        }


@pytest.mark.anyio
async def test_config_screen_can_disable_current_agent_skill() -> None:
    skill_item = SkillConfigItem(
        name="developer",
        description="Development workflow.",
        location="/skills/developer/SKILL.md",
    )
    screen = ConfigScreen(
        provider_name="openai",
        current_model="gpt-5.5",
        default_model="gpt-5.5",
        skills="auto",
        api_keys={"openai": "sk-test-ending"},
        skill_items=(skill_item,),
        skill_selection_mode="all",
        enabled_skills=(),
    )

    async with AceAITUI([]).run_test() as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()
        selections: list[ConfigSelection | None] = []

        def dismiss(selection: ConfigSelection | None) -> None:
            selections.append(selection)

        screen.dismiss = dismiss
        screen.query_one("#skill-0", Checkbox).value = False
        _press_config_apply(screen)

        assert selections[-1] == ConfigSelection(
            provider="openai",
            model="gpt-5.5",
            default_model="gpt-5.5",
            api_key="sk-test-ending",
            skills="auto",
            skill_selection_mode="selected",
            enabled_skills=(),
        )


@pytest.mark.anyio
async def test_config_screen_requires_provider_model_and_api_key() -> None:
    screen = ConfigScreen(
        provider_name="openai",
        current_model="gpt-5.5",
        default_model="gpt-5.5",
        skills="auto",
        api_keys={"openai": "sk-test-ending"},
    )

    async with AceAITUI([]).run_test() as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()
        selections: list[ConfigSelection | None] = []

        def dismiss(selection: ConfigSelection | None) -> None:
            selections.append(selection)

        screen.dismiss = dismiss
        provider_input = screen.query_one("#provider", Input)
        model_input = screen.query_one("#model", Input)
        api_key_input = screen.query_one("#api-key", Input)
        error = screen.query_one("#config-error", Static)

        provider_input.value = ""
        _press_config_apply(screen)
        assert selections == []
        assert str(error.render()) == "Provider is required"

        provider_input.value = "openai"
        model_input.value = ""
        _press_config_apply(screen)
        assert selections == []
        assert str(error.render()) == "Model is required"

        model_input.value = "gpt-5.5"
        api_key_input.value = ""
        _press_config_apply(screen)
        assert selections == []
        assert str(error.render()) == "API key is required"


@pytest.mark.anyio
async def test_config_screen_selects_prefixed_model_candidate_with_tab() -> None:
    screen = ConfigScreen(
        provider_name="deepseek",
        current_model="deepseek-v4-pro",
        default_model="deepseek-v4-pro",
        skills="auto",
        api_keys={},
    )

    async with AceAITUI([]).run_test() as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()
        model_input = screen.query_one("#model", Input)
        model_input.focus()

        model_input.value = "deepseek-c"
        await pilot.pause()
        await pilot.press("tab")

        assert model_input.value == "deepseek-chat"
        assert str(screen.query_one("#model-options", Static).render()) == ""


@pytest.mark.anyio
async def test_config_screen_selects_prefixed_provider_candidate_with_tab() -> None:
    screen = ConfigScreen(
        provider_name="openai",
        current_model="gpt-5.5",
        default_model="gpt-5.5",
        skills="auto",
        api_keys={"deepseek": "sk-deepseek-ending"},
    )

    async with AceAITUI([]).run_test() as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()
        provider_input = screen.query_one("#provider", Input)
        provider_input.focus()

        provider_input.value = "d"
        await pilot.pause()
        await pilot.press("tab")

        assert provider_input.value == "deepseek"
        assert screen.query_one("#model", Input).value == "deepseek-v4-pro"
        assert screen.query_one("#api-key", Input).value == "*****************ding"
        assert str(screen.query_one("#provider-options", Static).render()) == ""


@pytest.mark.anyio
async def test_config_screen_hides_candidates_for_empty_input() -> None:
    screen = ConfigScreen(
        provider_name="openai",
        current_model="gpt-5.5",
        default_model="gpt-5.5",
        skills="auto",
        api_keys={},
    )

    async with AceAITUI([]).run_test() as pilot:
        pilot.app.push_screen(screen)
        await pilot.pause()
        provider_input = screen.query_one("#provider", Input)
        provider_input.value = ""
        await pilot.pause()

        assert str(screen.query_one("#provider-options", Static).render()) == ""


@pytest.mark.anyio
async def test_configured_tui_switches_provider_without_reusing_current_key(
    tmp_path,
    monkeypatch,
) -> None:
    calls: list[AceAITUIConfig] = []

    def agent_factory(config: AceAITUIConfig) -> AgentBase:
        calls.append(config)
        return AgentBase(
            prompt="Prompt",
            default_model=config.model,
            llm_service=StubLLMService([]),  # type: ignore[arg-type]
            executor=StubExecutor(),  # type: ignore[arg-type]
        )

    monkeypatch.setenv("DEEPSEEK_API_KEY", "deepseek-key")
    monkeypatch.chdir(tmp_path)
    app = AceAIConfiguredTUI(
        agent_factory,
        initial_config=AceAITUIConfig(
            provider="openai",
            api_key="openai-key",
            model="gpt-5.5",
            api_keys={"openai": "openai-key"},
        ),
        initial_question="",
        default_model="gpt-5.5",
    )

    clear_config()
    async with app.run_test():
        app._handle_config_selection(
            ConfigSelection(
                provider="deepseek",
                model="deepseek-v4-flash",
                default_model="deepseek-v4-pro",
                api_key="",
                skills="auto",
                skill_selection_mode="selected",
                enabled_skills=(),
            )
        )

    assert calls[-1] == AceAITUIConfig(
        provider="deepseek",
        api_key="deepseek-key",
        model="deepseek-v4-flash",
        default_model="deepseek-v4-pro",
        skills="auto",
        skill_selection_mode="selected",
        enabled_skills=[],
        api_keys={"openai": "openai-key", "deepseek": "deepseek-key"},
    )
    assert current_config() == calls[-1]
    assert (tmp_path / ".aceai" / "config.yml").exists()


@pytest.mark.anyio
async def test_interactive_tui_session_selection_callback_switches_session(
    tmp_path,
) -> None:
    store = SessionStore(tmp_path)
    first = store.create_session()
    second = store.create_session()
    SessionRecorder(store, first.session_id).record(
        tui_event_to_session_event(TUIEvent.user_message("first"))
    )
    SessionRecorder(store, second.session_id).record(
        tui_event_to_session_event(TUIEvent.user_message("second"))
    )
    store.update_session_state(
        second.session_id,
        SessionState(selected_provider="openai", selected_model="gpt-5.5"),
    )
    llm_service = StubLLMService([])
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,  # type: ignore[arg-type]
        executor=StubExecutor(),  # type: ignore[arg-type]
    )
    app = AceAIInteractiveTUI(
        agent,
        initial_events=event_log_to_tui_events(store.load_event_log(first.session_id)),
        initial_history=[],
        session_recorder=SessionRecorder(store, first.session_id),
        session_id=first.session_id,
    )

    async with app.run_test():
        app._handle_session_selection(second.session_id)

        assert app._session_id == second.session_id
        assert app.title == f"AceAI {second.session_id}"
        assert app._state.events[0].content == "second"
        assert app._selected_model == "gpt-5.5"
        assert app._llm_history[0].content[0]["data"] == "second"


def _press_config_apply(screen: ConfigScreen) -> None:
    class Pressed:
        button = screen.query_one("#apply", Button)

    screen.on_button_pressed(Pressed())



"""Adapters from TUI events into durable AceAI session events."""

from typing import Any

from aceai.agent.session import SessionEvent

from .events import TUIEvent


def tui_event_to_session_event(event: TUIEvent) -> SessionEvent:
    return SessionEvent(
        run_id="",
        step_id=event.step_id,
        step_index=event.step_index,
        kind=event.kind,
        payload=_payload_for_event(event),
    )


def _payload_for_event(event: TUIEvent) -> dict[str, Any]:
    payload: dict[str, Any] = {"content": event.content}
    if event.tool_name is not None:
        payload["tool_name"] = event.tool_name
    if event.tool_call_id is not None:
        payload["tool_call_id"] = event.tool_call_id
    if event.tool_call is not None:
        payload["tool_call"] = event.tool_call.asdict()
    if event.tool_calls:
        payload["tool_calls"] = [call.asdict() for call in event.tool_calls]
    if event.tool_result is not None:
        payload["tool_result"] = {
            "output": event.tool_result.output,
            "error": event.tool_result.error,
        }
    if event.error is not None:
        payload["error"] = event.error
    if event.usage is not None:
        payload["usage"] = {
            "input_tokens": event.usage.input_tokens,
            "cached_input_tokens": event.usage.cached_input_tokens,
            "output_tokens": event.usage.output_tokens,
            "total_tokens": event.usage.total_tokens,
        }
    if event.cost is not None:
        payload["cost"] = event.cost.asdict()
    return payload



from aceai.agent.config import (
    AceAITUIConfig,
    LEGACY_AGENT_SKILLS_DIR,
    clear_config,
    config_schema,
    current_config,
    effective_config_path,
    load_config,
    project_config_path,
    replace_config,
    save_config,
)


def test_config_schema_lists_required_fields() -> None:
    schema = config_schema()

    fields = {field.name: field for field in schema.fields}

    assert fields["provider"].value_type == "string"
    assert fields["model"].required
    assert fields["default_model"].required
    assert fields["api_key"].required
    assert fields["skills"].value_type == "string"
    assert fields["skill_selection_mode"].value_type == "string"
    assert fields["enabled_skills"].value_type == "list"
    assert fields["api_keys"].value_type == "mapping"
    assert fields["tool_permissions"].value_type == "mapping"


def test_save_and_load_config_round_trips(tmp_path) -> None:
    path = tmp_path / "config.yaml"
    config = AceAITUIConfig(
        provider="openai",
        api_key="secret",
        model="gpt-4o-mini",
        api_keys={"openai": "secret"},
        tool_permissions={"run_shell_command": "never"},
    )

    save_config(config, path)
    loaded = load_config(path)

    assert loaded == config
    assert current_config() == config
    assert oct(path.stat().st_mode & 0o777) == "0o600"


def test_load_config_prefers_project_config_over_global(tmp_path, monkeypatch) -> None:
    monkeypatch.setenv("HOME", str(tmp_path / "home"))
    project_dir = tmp_path / "project"
    project_dir.mkdir()
    monkeypatch.chdir(project_dir)
    global_path = tmp_path / "home" / ".aceai" / "config.yaml"
    project_path = tmp_path / "project" / ".aceai" / "config.yml"
    global_path.parent.mkdir(parents=True)
    project_path.parent.mkdir(parents=True)
    save_config(
        AceAITUIConfig(
            provider="openai",
            api_key="global",
            model="gpt-4o-mini",
            api_keys={"openai": "global"},
        ),
        global_path,
    )
    project_config = AceAITUIConfig(
        provider="openai",
        api_key="project",
        model="gpt-5.5",
        api_keys={"openai": "project"},
    )
    save_config(project_config, project_path)

    assert effective_config_path() == project_path
    assert load_config() == project_config


def test_load_config_falls_back_to_global_config(tmp_path, monkeypatch) -> None:
    monkeypatch.setenv("HOME", str(tmp_path / "home"))
    project_dir = tmp_path / "project"
    project_dir.mkdir()
    monkeypatch.chdir(project_dir)
    global_path = tmp_path / "home" / ".aceai" / "config.yaml"
    global_config = AceAITUIConfig(
        provider="openai",
        api_key="global",
        model="gpt-4o-mini",
        api_keys={"openai": "global"},
    )
    save_config(global_config, global_path)

    assert effective_config_path() == global_path
    assert load_config() == global_config


def test_save_config_defaults_to_project_config(tmp_path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)
    config = AceAITUIConfig(
        provider="openai",
        api_key="secret",
        model="gpt-5.5",
        api_keys={"openai": "secret"},
    )

    save_config(config)

    path = project_config_path()
    assert path == tmp_path / ".aceai" / "config.yml"
    assert load_config(path) == config
    assert oct(path.stat().st_mode & 0o777) == "0o600"


def test_save_config_replaces_config_without_backup_files(tmp_path) -> None:
    path = tmp_path / "config.yaml"
    first = AceAITUIConfig(
        provider="openai",
        api_key="first",
        model="gpt-4o-mini",
        api_keys={"openai": "first"},
    )
    second = AceAITUIConfig(
        provider="openai",
        api_key="second",
        model="gpt-5.5",
        api_keys={"openai": "second"},
    )

    save_config(first, path)
    first_inode = path.stat().st_ino
    save_config(second, path)

    assert load_config(path) == second
    assert path.stat().st_ino != first_inode
    assert list(tmp_path.glob("*.bak")) == []
    assert list(tmp_path.glob(".config.yaml.*.tmp")) == []
    assert oct(path.stat().st_mode & 0o777) == "0o600"


def test_replace_config_uses_single_active_config_object() -> None:
    first = AceAITUIConfig(
        provider="openai",
        api_key="first",
        model="gpt-5.5",
        api_keys={"openai": "first"},
    )
    second = AceAITUIConfig(
        provider="openai",
        api_key="second",
        model="gpt-4o",
        api_keys={"openai": "second"},
    )

    replace_config(first)
    active_first = current_config()
    replace_config(second)

    assert active_first is first
    assert current_config() is second
    clear_config()
    assert current_config() is None


def test_load_config_migrates_stale_default_model(tmp_path) -> None:
    path = tmp_path / "config.yaml"
    path.write_text(
        "provider: openai\napi_keys:\n  openai: secret\nmodel: gpt-5.1\n",
        encoding="utf-8",
    )

    loaded = load_config(path)

    assert loaded == AceAITUIConfig(
        provider="openai",
        api_key="secret",
        model="gpt-5.5",
        api_keys={"openai": "secret"},
    )


def test_load_config_preserves_versioned_gpt51_selection(tmp_path) -> None:
    path = tmp_path / "config.yaml"
    config = AceAITUIConfig(
        provider="openai",
        api_key="secret",
        model="gpt-5.1",
        api_keys={"openai": "secret"},
    )
    save_config(config, path)

    loaded = load_config(path)

    assert loaded == config


def test_save_and_load_deepseek_config_round_trips(tmp_path) -> None:
    path = tmp_path / "config.yaml"
    config = AceAITUIConfig(
        provider="deepseek",
        api_key="secret",
        model="deepseek-v4-pro",
        default_model="deepseek-v4-pro",
        api_keys={"deepseek": "secret"},
    )

    save_config(config, path)
    loaded = load_config(path)

    assert loaded == config


def test_load_config_uses_provider_default_when_model_is_missing(tmp_path) -> None:
    path = tmp_path / "config.yaml"
    path.write_text(
        "config_version: 2\nprovider: deepseek\napi_keys:\n  deepseek: secret\n",
        encoding="utf-8",
    )

    loaded = load_config(path)

    assert loaded == AceAITUIConfig(
        provider="deepseek",
        api_key="secret",
        model="deepseek-v4-pro",
        default_model="deepseek-v4-pro",
        api_keys={"deepseek": "secret"},
    )


def test_load_config_replaces_legacy_builtin_skill_path_with_auto(tmp_path) -> None:
    path = tmp_path / "config.yaml"
    path.write_text(
        "config_version: 2\n"
        "provider: openai\n"
        "api_keys:\n"
        "  openai: secret\n"
        "model: gpt-5.5\n"
        "default_model: gpt-5.5\n"
        f"skills: {LEGACY_AGENT_SKILLS_DIR}\n",
        encoding="utf-8",
    )

    loaded = load_config(path)

    assert loaded == AceAITUIConfig(
        provider="openai",
        api_key="secret",
        model="gpt-5.5",
        default_model="gpt-5.5",
        skills="auto",
        api_keys={"openai": "secret"},
    )


def test_load_config_rejects_top_level_api_key(tmp_path) -> None:
    path = tmp_path / "config.yaml"
    path.write_text(
        "config_version: 2\nprovider: openai\napi_key: secret\n",
        encoding="utf-8",
    )

    try:
        load_config(path)
    except ValueError as exc:
        assert str(exc) == "AceAI config api_keys missing active provider"
    else:
        raise AssertionError("load_config accepted legacy top-level api_key")


def test_load_config_rejects_unknown_tool_permission(tmp_path) -> None:
    path = tmp_path / "config.yaml"
    path.write_text(
        "config_version: 2\n"
        "provider: openai\n"
        "api_keys:\n"
        "  openai: secret\n"
        "model: gpt-5.5\n"
        "tool_permissions:\n"
        "  run_shell_command: maybe\n",
        encoding="utf-8",
    )

    try:
        load_config(path)
    except ValueError as exc:
        assert str(exc) == "AceAI config tool_permissions value is unsupported"
    else:
        raise AssertionError("load_config accepted unknown tool permission")



"""Inline approval controls for suspended tool calls."""

import json

from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.events import Key
from textual.message import Message
from textual.widgets import Button, Static

from aceai.core.models import ToolApprovalRequest


class ApprovalWidget(Container):
    """Compact inline approval panel shown inside the chat surface."""

    class Selected(Message):
        def __init__(self, *, approved: bool) -> None:
            super().__init__()
            self.approved = approved

    DEFAULT_CSS = """
    ApprovalWidget {
        height: 3;
        background: #2e3440;
        border-top: solid #4c566a;
        padding: 0 1;
    }

    ApprovalWidget.collapsed {
        display: none;
    }

    #approval-row {
        height: 1;
        margin-top: 1;
    }

    #approval-summary {
        width: 1fr;
        height: 1;
        color: #d8dee9;
        margin-left: 1;
    }

    #approval-actions {
        width: auto;
        height: 1;
    }

    #approval-approve,
    #approval-reject {
        width: auto;
        min-width: 10;
        height: 1;
        padding: 0 1;
        border: none;
        background: transparent;
        text-style: bold;
    }

    #approval-approve {
        color: #a3be8c;
    }

    #approval-reject {
        color: #bf616a;
    }

    #approval-approve:focus,
    #approval-approve:hover {
        background: #3b4252;
        color: #a3be8c;
    }

    #approval-reject:focus,
    #approval-reject:hover {
        background: #3b4252;
        color: #bf616a;
    }
    """

    def compose(self) -> ComposeResult:
        with Horizontal(id="approval-row"):
            with Horizontal(id="approval-actions"):
                yield Button("A Approve", id="approval-approve")
                yield Button("R Reject", id="approval-reject")
            yield Static("", id="approval-summary")

    def show_request(self, request: ToolApprovalRequest) -> None:
        self.query_one("#approval-summary", Static).update(_approval_summary(request))
        self.remove_class("collapsed")
        self.query_one("#approval-approve", Button).focus()

    def clear_request(self) -> None:
        self.add_class("collapsed")
        self.query_one("#approval-summary", Static).update("")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "approval-approve":
            event.stop()
            self.post_message(self.Selected(approved=True))
            return
        if event.button.id == "approval-reject":
            event.stop()
            self.post_message(self.Selected(approved=False))

    def on_key(self, event: Key) -> None:
        if event.key == "a":
            event.stop()
            self.post_message(self.Selected(approved=True))
            return
        if event.key == "r":
            event.stop()
            self.post_message(self.Selected(approved=False))
            return
        if event.key == "right":
            event.stop()
            self.query_one("#approval-reject", Button).focus()
            return
        if event.key == "left":
            event.stop()
            self.query_one("#approval-approve", Button).focus()


def _approval_summary(request: ToolApprovalRequest) -> str:
    policy = request.policy or "approval required"
    arguments = _arguments_summary(request.call.arguments)
    return f"approval required  {request.tool_name}  {policy}  {arguments}"


def _arguments_summary(arguments: str) -> str:
    try:
        payload = json.loads(arguments)
    except json.JSONDecodeError:
        return _preview_line("arguments", arguments)
    if not isinstance(payload, dict):
        return _preview_line("arguments", arguments)
    lines: list[str] = []
    for key, value in payload.items():
        if isinstance(value, str):
            lines.append(_preview_line(key, value))
        elif isinstance(value, int | float | bool):
            lines.append(f"{key}: {value!r}")
        elif value is None:
            lines.append(f"{key}: null")
        else:
            lines.append(f"{key}: {value.__class__.__name__}")
    return "  ".join(lines)


def _preview_line(label: str, value: str) -> str:
    first_line = value.splitlines()[0] if value.splitlines() else ""
    if len(value) <= 96:
        return f"{label}: {value}"
    return f"{label}: {len(value)} chars, starts with {first_line[:72]}"



from pathlib import Path

import pytest

import scripts.release as release
from git import Repo


def test_read_current_version_does_not_exec_module(tmp_path: Path) -> None:
    version_file = tmp_path / "__init__.py"
    version_file.write_text(
        '__version__ = "1.2.3"\n\nfrom .core import AgentBase as AgentBase\n'
    )
    original_version_file = release.VERSION_FILE
    release.VERSION_FILE = version_file
    try:
        assert release.read_current_version() == "1.2.3"
    finally:
        release.VERSION_FILE = original_version_file


def test_read_current_version_requires_version_assignment(tmp_path: Path) -> None:
    version_file = tmp_path / "__init__.py"
    version_file.write_text('from .core import AgentBase as AgentBase\n')
    original_version_file = release.VERSION_FILE
    release.VERSION_FILE = version_file
    try:
        with pytest.raises(SystemExit, match="Could not locate __version__ assignment"):
            release.read_current_version()
    finally:
        release.VERSION_FILE = original_version_file


def test_read_latest_remote_tag_version_reads_from_origin(tmp_path: Path) -> None:
    origin_path = tmp_path / "origin.git"
    origin_repo = Repo.init(origin_path, bare=True)

    work_path = tmp_path / "work"
    repo = Repo.init(work_path)
    repo.create_remote("origin", str(origin_path))
    (work_path / "README.md").write_text("hi\n")
    repo.index.add(["README.md"])
    repo.index.commit("init")
    repo.create_tag("v0.1.5")
    repo.create_tag("v0.1.6")
    repo.git.push("--tags", "origin")

    assert release.read_latest_remote_tag_version(repo) == "0.1.6"


def test_ensure_version_order_compares_target_against_latest_remote_tag() -> None:
    release.ensure_version_order(
        "0.1.7",
        "0.1.7",
        "0.1.6",
        skip_update=False,
    )


def test_ensure_version_order_rejects_target_not_greater_than_remote_tag() -> None:
    with pytest.raises(SystemExit, match="Local branch version .* must be greater than latest remote tag"):
        release.ensure_version_order(
            "0.1.7",
            "0.1.6",
            "0.1.6",
            skip_update=False,
        )


def test_infer_version_from_version_branch(tmp_path: Path) -> None:
    repo = Repo.init(tmp_path)
    (tmp_path / "README.md").write_text("hi\n")
    repo.index.add(["README.md"])
    repo.index.commit("init")
    repo.git.checkout("-b", "version/1.2.3")

    assert release.infer_version_from_branch(repo) == "1.2.3"


def test_infer_version_requires_version_branch(tmp_path: Path) -> None:
    repo = Repo.init(tmp_path)
    (tmp_path / "README.md").write_text("hi\n")
    repo.index.add(["README.md"])
    repo.index.commit("init")
    repo.git.checkout("-b", "feature/foo")

    with pytest.raises(SystemExit, match="Provide --version or check out version/"):
        release.infer_version_from_branch(repo)



from typing import Any, Literal

from msgspec import Struct, field

from aceai.llm.models import LLMToolCall

from .events import AgentEventBuilder
from .models import AgentStep, ToolApprovalRequest
from .tools import Tool

AgentRuntimeStatus = Literal["running", "suspended", "completed", "failed"]


class ToolInvocation(Struct, kw_only=True):
    """Resolved local tool call ready for policy checks and execution."""

    call: LLMToolCall
    tool: Tool[Any, Any]

    @property
    def approval_required(self) -> bool:
        return self.tool.metadata.require_approval


class ToolRunState(Struct, kw_only=True):
    """Tool execution bookkeeping for a single agent run."""

    call_counts: dict[str, int] = field(default_factory=dict[str, int])
    approved_tool_names: set[str] = field(default_factory=set[str])


class PendingToolApproval(Struct, kw_only=True):
    """Suspended continuation waiting for a caller approval decision."""

    step: AgentStep
    invocation: ToolInvocation
    request: ToolApprovalRequest
    event_builder: AgentEventBuilder
    tool_index: int


class AgentRuntimeState(Struct, kw_only=True):
    """Mutable runtime state for one AgentRuntime."""

    status: AgentRuntimeStatus = "running"
    tools: ToolRunState = field(default_factory=ToolRunState)
    pending_approval: PendingToolApproval | None = None

    def suspend_for_approval(
        self,
        *,
        step: AgentStep,
        invocation: ToolInvocation,
        request: ToolApprovalRequest,
        event_builder: AgentEventBuilder,
        tool_index: int,
    ) -> None:
        self.status = "suspended"
        self.pending_approval = PendingToolApproval(
            step=step,
            invocation=invocation,
            request=request,
            event_builder=event_builder,
            tool_index=tool_index,
        )

    def resume_from_approval(self) -> PendingToolApproval:
        pending = self.pending_approval
        if pending is None:
            raise ValueError("agent run is not suspended for tool approval")
        self.pending_approval = None
        self.status = "running"
        return pending



"""Compatibility exports for AceAI TUI configuration."""

from aceai.agent.config import (
    APP_CONFIG_SCHEMA as APP_CONFIG_SCHEMA,
    CONFIG_VERSION as CONFIG_VERSION,
    DEFAULT_OPENAI_MODEL as DEFAULT_OPENAI_MODEL,
    OPENAI_MODEL_OPTIONS as OPENAI_MODEL_OPTIONS,
    STALE_DEFAULT_OPENAI_MODELS as STALE_DEFAULT_OPENAI_MODELS,
    SUPPORTED_OPENAI_MODELS as SUPPORTED_OPENAI_MODELS,
    AceAITUIConfig as AceAITUIConfig,
    AgentAppConfig as AgentAppConfig,
    ConfigField as ConfigField,
    ConfigSchema as ConfigSchema,
    ToolPermission as ToolPermission,
    clear_config as clear_config,
    config_schema as config_schema,
    current_config as current_config,
    default_config_path as default_config_path,
    effective_config_path as effective_config_path,
    load_config as load_config,
    project_config_path as project_config_path,
    replace_config as replace_config,
    save_config as save_config,
    validate_config as validate_config,
)



from typing import Annotated as Annotated

from ._tool_sig import spec as spec
from .tool import IToolSpec as IToolSpec
from .tool import OpenAIToolSpec as OpenAIToolSpec
from .tool import Tool as Tool
from .tool import ToolMeta as ToolMeta
from .tool import tool as tool



import pytest

from aceai.agent.app import AceAgentApp
from aceai.agent.session import SessionEvent, SessionStore
from aceai.core.base import AgentBase
from aceai.core.events import RunSuspendedEvent, ToolApprovalRequestedEvent
from aceai.llm import LLMResponse
from aceai.llm.models import LLMStreamEvent, LLMToolCall

from tests.test_agent_behavior import StubExecutor, StubLLMService, make_stream


@pytest.mark.anyio
async def test_agent_app_reuses_approved_tool_name_across_session_turns(tmp_path) -> None:
    first_call = LLMToolCall(
        name="write_file",
        arguments='{"path":"a"}',
        call_id="write-1",
    )
    second_call = LLMToolCall(
        name="write_file",
        arguments='{"path":"b"}',
        call_id="write-2",
    )
    llm_service = StubLLMService(
        [
            [
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="use write", tool_calls=[first_call]),
                )
            ],
            make_stream(response=LLMResponse(text="first done"), deltas=["first done"]),
            [
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="use write", tool_calls=[second_call]),
                )
            ],
            make_stream(response=LLMResponse(text="second done"), deltas=["second done"]),
        ]
    )
    executor = StubExecutor(
        {"write_file": '{"ok":true}'},
        approval_required={"write_file"},
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,
        executor=executor,
        max_steps=2,
    )
    app = AceAgentApp(
        agent,
        provider_name="openai",
        selected_model="gpt-4o",
        session_store=SessionStore(tmp_path / "sessions"),
    )

    first_events = [event async for event in app.start_turn("Write one")]
    assert [event for event in first_events if isinstance(event, RunSuspendedEvent)]
    [event async for event in app.approve_tool()]

    second_events = [event async for event in app.start_turn("Write another")]

    assert executor.calls == [first_call, second_call]
    assert not [
        event for event in second_events if isinstance(event, ToolApprovalRequestedEvent)
    ]
    assert not [event for event in second_events if isinstance(event, RunSuspendedEvent)]


@pytest.mark.anyio
async def test_agent_app_seeds_approved_tool_cache_from_session_history(
    tmp_path,
) -> None:
    call = LLMToolCall(
        name="write_file",
        arguments='{"path":"b"}',
        call_id="write-2",
    )
    store = SessionStore(tmp_path / "sessions")
    metadata = store.create_session()
    store.append_event(
        metadata.session_id,
        SessionEvent(
            kind="tool_approval_resolved",
            payload={
                "content": "approved",
                "tool_name": "write_file",
                "tool_call_id": "write-1",
                "tool_call": {
                    "type": "function_call",
                    "name": "write_file",
                    "arguments": '{"path":"a"}',
                    "call_id": "write-1",
                },
            },
        ),
    )
    llm_service = StubLLMService(
        [
            [
                LLMStreamEvent(
                    event_type="response.completed",
                    response=LLMResponse(text="use write", tool_calls=[call]),
                )
            ],
            make_stream(response=LLMResponse(text="done"), deltas=["done"]),
        ]
    )
    executor = StubExecutor(
        {"write_file": '{"ok":true}'},
        approval_required={"write_file"},
    )
    agent = AgentBase(
        prompt="Prompt",
        default_model="gpt-4o",
        llm_service=llm_service,
        executor=executor,
        max_steps=2,
    )
    app = AceAgentApp(
        agent,
        provider_name="openai",
        selected_model="gpt-4o",
        session_store=store,
        session_id=metadata.session_id,
    )

    events = [event async for event in app.start_turn("Write another")]

    assert executor.calls == [call]
    assert not [event for event in events if isinstance(event, ToolApprovalRequestedEvent)]
    assert not [event for event in events if isinstance(event, RunSuspendedEvent)]



from typing import AsyncGenerator

import httpx
import pytest
from ididi import Graph, use
from msgspec import Struct

from aceai.llm.errors import AceAIRuntimeError
from aceai.core.executor import LoggingToolExecutor, ToolExecutor
from aceai.core.run_state import ToolRunState
from aceai.llm.models import LLMToolCall
from aceai.core.tools import tool
from aceai.core.tools._tool_sig import Annotated, spec


class FakeLogger:
    def __init__(self):
        self.info_messages: list[str] = []
        self.success_messages: list[str] = []
        self.exception_messages: list[str] = []

    def info(self, msg: str, /, **_: object) -> None:
        self.info_messages.append(msg)

    def success(self, msg: str, /, **_: object) -> None:
        self.success_messages.append(msg)

    def exception(self, msg: str, /, **_: object) -> None:
        self.exception_messages.append(msg)


class StepTimer:
    def __init__(self, *ticks: float):
        self._ticks = iter(ticks)

    def __call__(self) -> float:
        return next(self._ticks)


class UserRepo:
    def __init__(self, label: str):
        self.label = label


def provide_user_repo() -> UserRepo:
    return UserRepo("primary")


def describe_user(
    repo: Annotated[UserRepo, use(provide_user_repo)],
    user_id: Annotated[int, spec(description="User identifier")],
) -> str:
    return f"{repo.label}:{user_id}"


async def async_increment(
    value: Annotated[int, spec(description="Value to increment")],
) -> int:
    return value + 1


def unreliable_tool(
    value: Annotated[int, spec(description="Value that triggers failure")],
) -> int:
    raise AceAIRuntimeError("expected failure")


def echo_message(
    message: Annotated[str, spec(description="Message to echo")],
) -> str:
    return message


def build_async_client() -> httpx.AsyncClient:
    return httpx.AsyncClient()


async def identify_httpx_client(
    client: Annotated[httpx.AsyncClient, use(build_async_client)],
) -> str:
    try:
        return client.__class__.__name__
    finally:
        await client.aclose()


def build_tool(func):
    return tool(func)


async def execute_call(
    executor: ToolExecutor,
    call: LLMToolCall,
    tool_state: ToolRunState | None = None,
) -> str:
    return await executor.execute(
        executor.resolve_invocation(call),
        tool_state=tool_state or ToolRunState(),
    )


@pytest.fixture
async def graph() -> AsyncGenerator[Graph, None]:
    graph = Graph()
    try:
        yield graph
    finally:
        graph._workers.shutdown(wait=True)


@pytest.mark.anyio
async def test_tool_executor_executes_tool_with_dep_graph(graph: Graph) -> None:
    describe_tool = build_tool(describe_user)
    executor = ToolExecutor(graph, [describe_tool])

    call = LLMToolCall(
        name=describe_tool.name,
        arguments='{"user_id":7}',
        call_id="call-123",
    )

    encoded_result = await execute_call(executor, call)

    assert encoded_result == '"primary:7"'


@pytest.mark.anyio
async def test_tool_executor_awaits_async_tool_results(graph: Graph) -> None:
    increment_tool = build_tool(async_increment)
    executor = ToolExecutor(graph, [increment_tool])

    call = LLMToolCall(
        name=increment_tool.name,
        arguments='{"value":2}',
        call_id="call-async",
    )

    result = await execute_call(executor, call)

    assert result == "3"


@pytest.mark.anyio
async def test_logging_tool_executor_logs_successful_calls(graph: Graph) -> None:
    increment_tool = build_tool(async_increment)
    logger = FakeLogger()
    timer = StepTimer(10.0, 10.5)
    executor = LoggingToolExecutor(
        graph=graph,
        tools=[increment_tool],
        logger=logger,
        timer=timer,
    )

    call = LLMToolCall(
        name=increment_tool.name,
        arguments='{"value":2}',
        call_id="req-success",
    )

    result = await execute_call(executor, call)

    assert result == "3"
    assert logger.info_messages == [
        'Tool async_increment starting (call_id=req-success) with {"value":2}'
    ]
    assert logger.success_messages == [
        "Tool async_increment finished in 0.50s, result: 3"
    ]
    assert logger.exception_messages == []


@pytest.mark.anyio
async def test_logging_tool_executor_logs_and_reraises_failures(graph: Graph) -> None:
    failing_tool = build_tool(unreliable_tool)
    logger = FakeLogger()
    timer = StepTimer(5.0, 6.25)
    executor = LoggingToolExecutor(
        graph=graph,
        tools=[failing_tool],
        logger=logger,
        timer=timer,
    )

    call = LLMToolCall(
        name=failing_tool.name,
        arguments='{"value":1}',
        call_id="req-fail",
    )

    with pytest.raises(AceAIRuntimeError, match="expected failure"):
        await execute_call(executor, call)

    assert logger.info_messages == [
        'Tool unreliable_tool starting (call_id=req-fail) with {"value":1}'
    ]
    assert logger.exception_messages == ["Tool unreliable_tool failed after 1.25s"]
    assert logger.success_messages == []


@pytest.mark.anyio
async def test_tool_executor_resolves_httpx_async_client(graph: Graph) -> None:
    client_tool = build_tool(identify_httpx_client)
    executor = ToolExecutor(graph, [client_tool])

    call = LLMToolCall(name=client_tool.name, arguments="{}", call_id="req-httpx")

    result = await execute_call(executor, call)

    assert result == '"AsyncClient"'


class Payload(Struct):
    value: int


def build_struct(value: Annotated[int, spec(description="Value to wrap")]) -> Payload:
    return Payload(value)


@pytest.mark.anyio
async def test_tool_executor_encodes_struct_return(graph: Graph) -> None:
    struct_tool = build_tool(build_struct)
    executor = ToolExecutor(graph, [struct_tool])

    call = LLMToolCall(
        name=struct_tool.name,
        arguments='{"value":5}',
        call_id="req-struct",
    )

    encoded = await execute_call(executor, call)

    assert encoded == '{"value":5}'


@pytest.mark.anyio
async def test_tool_executor_exposes_tool_specs(graph: Graph) -> None:
    echo_tool = build_tool(echo_message)
    executor = ToolExecutor(graph, [echo_tool])

    first = executor.all_tools
    second = executor.all_tools

    assert first is second
    schema_names = {spec.name for spec in first}
    assert echo_tool.name in schema_names


@pytest.mark.anyio
async def test_tool_executor_enforces_max_calls_per_run(graph: Graph) -> None:
    executed: list[int] = []

    def tick() -> int:
        executed.append(1)
        return len(executed)

    tick_tool = tool(max_calls_per_run=1)(tick)
    executor = ToolExecutor(graph, [tick_tool])
    tool_state = ToolRunState()

    first = await execute_call(
        executor,
        LLMToolCall(name=tick_tool.name, arguments="{}", call_id="call-1"),
        tool_state,
    )
    second = await execute_call(
        executor,
        LLMToolCall(name=tick_tool.name, arguments="{}", call_id="call-2"),
        tool_state,
    )

    assert first == "1"
    assert second == (
        "the tool tick exceeds its max calls in this run, do not call it again"
    )
    assert executed == [1]


@pytest.mark.anyio
async def test_tool_executor_resolves_invocation_with_approval_metadata(
    graph: Graph,
) -> None:
    approved_tool = tool(require_approval=True)(echo_message)
    executor = ToolExecutor(graph, [approved_tool])
    call = LLMToolCall(
        name=approved_tool.name,
        arguments='{"message":"hello"}',
        call_id="call-approval",
    )

    invocation = executor.resolve_invocation(call)
    result = await executor.execute(invocation, tool_state=ToolRunState())

    assert invocation.call is call
    assert invocation.tool is approved_tool
    assert invocation.approval_required is True
    assert result == '"hello"'



"""DeepSeek provider using the OpenAI-compatible Chat Completions API."""

import time
from typing import Any, AsyncGenerator

from openai import AsyncOpenAI

from aceai.llm.errors import AceAIConfigurationError
from aceai.llm.interface import UNSET, Unset
from aceai.llm.models import (
    LLMHostedToolSpec,
    LLMInput,
    LLMMessage,
    LLMProviderModality,
    LLMResponse,
    LLMSegment,
    LLMStreamEvent,
    LLMToolCall,
    LLMToolCallDelta,
    LLMToolCallMessage,
    LLMToolCallSegmentMeta,
    LLMToolSpec,
    LLMToolUseMessage,
    LLMUsage,
    LLMReasoningSegmentMeta,
)
from aceai.llm.openai import OpenAI, OpenAIMeta, OpenAIPayload


DEEPSEEK_BASE_URL = "https://api.deepseek.com"


class DeepSeek(OpenAI):
    """DeepSeek provider backed by OpenAI-compatible chat completions."""

    def __init__(
        self,
        *,
        api_key: str,
        default_meta: OpenAIMeta,
        base_url: str = DEEPSEEK_BASE_URL,
    ):
        super().__init__(
            client=AsyncOpenAI(api_key=api_key, base_url=base_url),
            default_meta=default_meta,
            provider_name="deepseek",
        )

    @property
    def modality(self) -> LLMProviderModality:
        return LLMProviderModality()

    def _coerce_text_content(
        self,
        content: list[dict[str, Any]],
        *,
        context: str,
    ) -> str:
        text_parts: list[str] = []
        for part in content:
            if part["type"] != "text":
                raise ValueError(f"{context} only supports text parts")
            data = part["data"]
            if type(data) is not str:
                raise TypeError(f"{context} text parts must be str")
            text_parts.append(data)
        return "".join(text_parts)

    def _format_messages_for_chat(
        self, messages: list[LLMMessage]
    ) -> list[dict[str, Any]]:
        formatted: list[dict[str, Any]] = []
        for message in messages:
            if isinstance(message, LLMToolUseMessage):
                formatted.append(
                    {
                        "role": "tool",
                        "tool_call_id": message.call_id,
                        "name": message.name,
                        "content": self._coerce_text_content(
                            message.content,
                            context="DeepSeek tool output",
                        ),
                    }
                )
                continue
            if isinstance(message, LLMToolCallMessage):
                assistant_message = {
                    "role": "assistant",
                    "content": self._coerce_text_content(
                        message.content,
                        context="DeepSeek assistant message",
                    ),
                    "tool_calls": [
                        self._format_tool_call_message(call)
                        for call in message.tool_calls
                    ],
                }
                if message.reasoning_content is not None:
                    assistant_message["reasoning_content"] = message.reasoning_content
                formatted.append(assistant_message)
                continue
            formatted.append(
                {
                    "role": message.role,
                    "content": self._coerce_text_content(
                        message.content,
                        context="DeepSeek message",
                    ),
                }
            )
        return formatted

    def _format_tool_call_message(self, call: LLMToolCall) -> dict[str, Any]:
        return {
            "id": call.call_id,
            "type": "function",
            "function": {
                "name": call.name,
                "arguments": call.arguments,
            },
        }

    def _format_tool_for_chat(self, tool: LLMToolSpec) -> dict[str, Any]:
        if isinstance(tool, LLMHostedToolSpec):
            raise AceAIConfigurationError("DeepSeek does not support hosted tools")
        schema = tool.generate_schema()
        return {
            "type": "function",
            "function": {
                "name": schema["name"],
                "description": schema["description"],
                "parameters": schema["parameters"],
                "strict": schema["strict"],
            },
        }

    def _build_chat_kwargs(self, payload: OpenAIPayload) -> dict[str, Any]:
        data = payload.asdict()
        metadata = payload.metadata if "metadata" in data else {}
        if "model" not in metadata:
            raise AceAIConfigurationError(
                "DeepSeek request metadata must include a model identifier"
            )
        model_name = metadata["model"]

        kwargs: dict[str, Any] = {
            "model": model_name,
            "messages": self._format_messages_for_chat(payload.messages),
        }
        if "temperature" in data:
            kwargs["temperature"] = payload.temperature
        if "top_p" in data:
            kwargs["top_p"] = payload.top_p
        if "max_tokens" in data:
            kwargs["max_tokens"] = payload.max_tokens
        if "stop" in data:
            kwargs["stop"] = payload.stop
        if "tools" in data:
            kwargs["tools"] = [
                self._format_tool_for_chat(tool) for tool in payload.tools
            ]
        if "tool_choice" in data:
            kwargs["tool_choice"] = payload.tool_choice
        if "response_format" in data:
            kwargs["response_format"] = {"type": payload.response_format.type}
        return kwargs

    def _usage_from_chat(self, usage: Any) -> Unset[LLMUsage]:
        if usage is None:
            return UNSET
        cached_input_tokens: int | None = None
        if usage.prompt_tokens_details is not None:
            cached_input_tokens = usage.prompt_tokens_details.cached_tokens
        return LLMUsage(
            input_tokens=usage.prompt_tokens,
            cached_input_tokens=cached_input_tokens,
            output_tokens=usage.completion_tokens,
            total_tokens=usage.total_tokens,
        )

    def _reasoning_content_from_extra(self, value: Any) -> str | None:
        extra = value.model_extra
        if extra is None:
            return None
        if "reasoning_content" not in extra:
            return None
        reasoning_content = extra["reasoning_content"]
        if reasoning_content is None:
            return None
        if type(reasoning_content) is not str:
            raise TypeError("DeepSeek reasoning_content must be str")
        return reasoning_content

    def _tool_calls_from_chat(self, message: Any) -> list[LLMToolCall]:
        calls: list[LLMToolCall] = []
        if message.tool_calls is None:
            return calls
        for call in message.tool_calls:
            calls.append(
                LLMToolCall(
                    name=call.function.name,
                    arguments=call.function.arguments,
                    call_id=call.id,
                )
            )
        return calls

    def _to_chat_llm_response(
        self,
        response: Any,
        *,
        latency_ms: float | None = None,
    ) -> LLMResponse:
        choice = response.choices[0]
        message = choice.message
        content = message.content
        if content is None:
            content = ""
        reasoning_content = self._reasoning_content_from_extra(message)
        tool_calls = self._tool_calls_from_chat(message)
        segments = [LLMSegment(type="text", content=content)] if content != "" else []
        if reasoning_content is not None:
            segments.append(
                LLMSegment(
                    type="reasoning",
                    content=reasoning_content,
                )
            )
        for call in tool_calls:
            segments.append(
                LLMSegment(
                    type="tool_call",
                    content=call.arguments,
                    meta=LLMToolCallSegmentMeta(
                        call_id=call.call_id,
                        tool_name=call.name,
                    ),
                )
            )
        return LLMResponse(
            id=response.id,
            model=response.model,
            text=content,
            tool_calls=tool_calls,
            usage=self._usage_from_chat(response.usage),
            segments=segments,
            provider_meta=[
                self._provider_meta_entry(
                    model=response.model,
                    latency_ms=latency_ms,
                    response_id=response.id,
                )
            ],
            status=choice.finish_reason,
            reasoning_content=reasoning_content,
        )

    def request_to_payload(self, request: LLMInput) -> OpenAIPayload:
        payload = OpenAIPayload.from_input(request)
        return self._apply_default_meta(payload)

    async def complete(self, request: LLMInput) -> LLMResponse:
        payload = self.request_to_payload(request)
        kwargs = self._build_chat_kwargs(payload)
        start = time.perf_counter()
        response = await self._client.chat.completions.create(**kwargs)
        latency_ms = (time.perf_counter() - start) * 1000.0
        return self._to_chat_llm_response(response, latency_ms=latency_ms)

    async def stream(self, request: LLMInput) -> AsyncGenerator[LLMStreamEvent, None]:
        payload = self.request_to_payload(request)
        kwargs = self._build_chat_kwargs(payload)
        kwargs["stream"] = True
        kwargs["stream_options"] = {"include_usage": True}
        start = time.perf_counter()
        stream = await self._client.chat.completions.create(**kwargs)
        text = ""
        reasoning_content = ""
        model_name = kwargs["model"]
        usage: Unset[LLMUsage] = UNSET
        tool_names_by_index: dict[int, str] = {}
        tool_ids_by_index: dict[int, str] = {}
        tool_arguments_by_index: dict[int, str] = {}

        async for chunk in stream:
            if chunk.usage is not None:
                usage = self._usage_from_chat(chunk.usage)
            if len(chunk.choices) == 0:
                continue
            choice = chunk.choices[0]
            delta = choice.delta
            delta_reasoning = self._reasoning_content_from_extra(delta)
            if delta_reasoning is not None:
                reasoning_content += delta_reasoning
                yield LLMStreamEvent(
                    event_type="response.reasoning.delta",
                    segments=[
                        LLMSegment(
                            type="reasoning",
                            content=delta_reasoning,
                            meta=LLMReasoningSegmentMeta(
                                item_id="deepseek-reasoning",
                                kind="content",
                                index=0,
                                status="in_progress",
                                is_delta=True,
                            ),
                        )
                    ],
                    provider_meta=[self._provider_meta_entry(model=model_name)],
                )
                continue
            if delta.content:
                text += delta.content
                yield LLMStreamEvent(
                    event_type="response.output_text.delta",
                    text_delta=delta.content,
                    segments=[LLMSegment(type="text", content=delta.content)],
                    provider_meta=[self._provider_meta_entry(model=model_name)],
                )
            if delta.tool_calls is None:
                continue
            for tool_delta in delta.tool_calls:
                if tool_delta.id is not None:
                    tool_ids_by_index[tool_delta.index] = tool_delta.id
                if tool_delta.function.name is not None:
                    tool_names_by_index[tool_delta.index] = tool_delta.function.name
                if tool_delta.function.arguments is None:
                    continue
                existing = (
                    tool_arguments_by_index[tool_delta.index]
                    if tool_delta.index in tool_arguments_by_index
                    else ""
                )
                tool_arguments_by_index[tool_delta.index] = (
                    existing + tool_delta.function.arguments
                )
                call_id = tool_ids_by_index[tool_delta.index]
                yield LLMStreamEvent(
                    event_type="response.function_call_arguments.delta",
                    tool_call_delta=LLMToolCallDelta(
                        id=call_id,
                        arguments_delta=tool_delta.function.arguments,
                    ),
                    segments=[
                        LLMSegment(
                            type="tool_call",
                            content=tool_delta.function.arguments,
                            meta=LLMToolCallSegmentMeta(
                                call_id=call_id,
                                tool_name=(
                                    tool_names_by_index[tool_delta.index]
                                    if tool_delta.index in tool_names_by_index
                                    else None
                                ),
                                is_delta=True,
                            ),
                        )
                    ],
                    provider_meta=[self._provider_meta_entry(model=model_name)],
                )

        latency_ms = (time.perf_counter() - start) * 1000.0
        tool_calls = [
            LLMToolCall(
                name=tool_names_by_index[index],
                arguments=tool_arguments_by_index[index],
                call_id=tool_ids_by_index[index],
            )
            for index in sorted(tool_arguments_by_index)
        ]
        segments = [LLMSegment(type="text", content=text)] if text != "" else []
        if reasoning_content != "":
            segments.append(
                LLMSegment(
                    type="reasoning",
                    content=reasoning_content,
                )
            )
        for call in tool_calls:
            segments.append(
                LLMSegment(
                    type="tool_call",
                    content=call.arguments,
                    meta=LLMToolCallSegmentMeta(
                        call_id=call.call_id,
                        tool_name=call.name,
                    ),
                )
            )
        final_response = LLMResponse(
            model=model_name,
            text=text,
            tool_calls=tool_calls,
            usage=usage,
            segments=segments,
            provider_meta=[
                self._provider_meta_entry(
                    model=model_name,
                    latency_ms=latency_ms,
                )
            ],
            status="completed",
            reasoning_content=reasoning_content if reasoning_content != "" else None,
        )
        yield LLMStreamEvent(
            event_type="response.completed",
            response=final_response,
            segments=final_response.segments,
            provider_meta=final_response.provider_meta,
        )
